""" WildCard Import
"""
from pyrusgeom.angle_deg import *
from pyrusgeom.vector_2d import *
from pyrusgeom.line_2d import *
from pyrusgeom.ray_2d import *
from pyrusgeom.sector_2d import *
from pyrusgeom.triangle_2d import Triangle2D
from pyrusgeom.segment_2d import Segment2D
from pyrusgeom.matrix_2d import *  # not necessary but optimized for 2d
from pyrusgeom.circle_2d import *
from pyrusgeom.rect_2d import *
from pyrusgeom.polygon_2d import *
